#!/usr/local/bin/python
#
# hourly_tec_vs_daily.py -- Compare hourly TEC files to daily files
#                           to ensure that TEC values are correct.
#

USAGE = """
hourly_tec_vs_daily.py -d <daily tec_dump file> [-o <augmented hourly TEC1 lines>] [-s <site statistics file>]
                       [-t <max_time_diff] [-v <gps_to_prn table file>] <hourly TEC1 files> ...

Switches:
  --help,         -h  = print usage
  --daily   <fn>, -d  = input  file of daily TEC lines in tec_dump format
 [--output  <fn>, -o] = output file of augmented TEC lines comparing hourly to daily
 [--stats   <fn>, -s] = statistics output file
 [--time_max_diff <sec>, -t = max. allowable time difference in match ups (default 15 sec.)
  --svn2prn <fn>, -v  = svn_to_prn lookup table file


"""

import sys, os, re, string, getopt, bisect
from sys import argv
from string import split, join
def echo2(str): sys.stderr.write(str + "\n")
def die(str, status=1): echo2(str); sys.exit(status)

tec1_file = ""; tec2_file = ""; output_file = ""; stats_file = ""; svn2prn_file = ""
max_time_delta = 120.
me = argv[0]
try:
    opts, argv = getopt.getopt(argv[1:], 'hd:o:s:t:v:',
                  ['help', 'daily=', 'output=', 'stats=', 'time_max_diff=', 'svn2prn='])
except getopt.GetoptError, (msg, bad_opt):
    die("%s error: Bad option: %s, %s" % (me, bad_opt, msg))
for opt, val in opts:
    if opt in ("-h", "--help"):     die(USAGE)
    elif opt in ("-d", "--daily"):  tec1_file  = val
    elif opt in ("-o", "--output"): output_file = val
    elif opt in ("-s", "--stats"):  stats_file = val
    elif opt in ("-t", "--time_max_diff"):  max_time_delta = val
    elif opt in ("-v", "--svn2prn"):  svn2prn_file = val    

if tec1_file == "": die(USAGE)
if svn2prn_file == "": die(USAGE)
if len(argv) < 1: die(USAGE)
tec2_file = argv[0]
if tec2_file == "": die(USAGE)
if stats_file != "":
    keys = [(2, 'site'), (3, 'sat'), (1, 'time', 'numeric')]
    vars = [(23, 'vtec_diff'), (21, 'vtec')]
    config = {}
    gstats = GroupedStats(keys, vars, config)

class SVN2PRN:
    """SVN to PRN translation table"""
    def __init__(self, file):
        self.svn2prn = {}
        self.prn2svn = {}
        for line in open(file, "r"):
            fields = split(line)
            svn = "%2.2d" % int(fields[0])
            prn = "%2.2d" % int(fields[1])
            self.svn2prn[svn] = prn
            self.prn2svn[prn] = svn
tbl = SVN2PRN(svn2prn_file)
            
class TECPt:
    """Holder class for TEC fields: time,site,sat,az,el,lat,lon,tec,err,date_time"""
    def __init__(self, line, mode=1):
        fields = split(line)
        if ( mode == 1 ):
            init = map(lambda i: fields[i], [30,1,2,3,4,5,6,14,18,31])
            # column numbers for fields in tec_dump file
        elif ( mode == 2 ):
            init = map(lambda i: fields[i], [0,1,2,7,8,9,10,3,4,11])
            # column numbers for fields in hourly tec_vs_time_afrl file
            init[1] = init[1].upper()                # upper case site
            init[2] = "GPS" + tbl.prn2svn[init[2]]   # sat = GPS## where ## = SVN
        else:
            die("hourly_tec_vs_daily.py: Error, bad TEC fields online.")
        self.time = float(init[0])
        self.site = init[1]; self.sat = init[2];
        self.az = init[3]; self.el = init[4];
        self.lat = init[5]; self.lon = init[6];
        self.val = float(init[7]); self.err = init[8];
        self.date_time = init[9]

# Sort TEC values into site/sat/time order for matchup.
# Aux. lists contain 3 sort keys before TECPt itself.
def TECPt_Sort_Augment1(x):
    fields = split(x)
    return (fields[1], fields[2], float(fields[30]), x)

def TECPt_Sort_Augment2(x):
    fields = split(x)
    return (fields[1].upper(), "GPS" + tbl.prn2svn[fields[2]], float(fields[0]), x)

def match_up_TEC_obs(aux_tecs1, aux_tec2):
    i1 = bisect.bisect(aux_tecs1, aux_tec2)
    tec1_line = aux_tecs1[i1][3]
    tec1 = TECPt(tec1_line)
    tec2_line = aux_tec2[3]
    tec2 = TECPt(tec2_line, mode=2)
    
    if ( tec1.site == tec2.site and tec1.sat == tec2.sat):
        match = 1
        time_diff = tec2.time - tec1.time
        for i in [-1, 1]:
            tec1_try_line = aux_tecs1[i1+i][3]
            tec1_try = TECPt(tec1_try_line)
            if ( tec1_try.site == tec2.site and tec1_try.sat == tec2.sat):
                time_diff_try = tec2.time - tec1_try.time
                if ( abs(time_diff_try) < abs(time_diff) ):
                    time_diff = time_diff_try
                    tec1 = tec1_try
    else:
        match = 0
        time_diff = 999999.
    return (match, time_diff, tec1, tec2)


# main
echo2("Reading 1st TEC file . . .")
tecs1 = []
for line in open(tec1_file, "r"):
    if (line[0] == '#'): continue
#    fields = split(line)
#    tec_fields = map(lambda i: fields[i], [0,1,2,3,4,5,6,14,18,12])
#    pt = TECPt(tec_fields)
#    pt.err = float(fields[18]) / float(fields[19])    
#    tecs1.append(pt)
    tecs1.append(line)

echo2("Sorting 1st TEC lines . . .")
aux_tecs1 = map(TECPt_Sort_Augment1, tecs1)
aux_tecs1.sort()

echo2("Reading 2nd TEC file(s) . . .")
print "#Time Site Sat VTEC Err STEC Err Az El Lat Lon DateTime SiteLat SiteLon ", \
      "TimeDiff SVN Az1 El1 Lat1 Lon1 VTEC1 Err1 VTEC_Diff"
tecs2 = []
in2 = open(tec2_file, "r")
line = in2.readline()
while (line and line[0] == '#'): line = in2.readline()
fields = split(line)
cur_site = fields[1]
while line:
    fields = split(line)
    if fields[1] == cur_site:
        # Read blocks of input lines that are from same site
        tecs2.append(line)
    else:
        # Sort site block by site/sat/time and then match up for comparison
        aux_tecs2 = map(TECPt_Sort_Augment2, tecs2)
        aux_tecs2.sort()
        for i2 in xrange( len(aux_tecs2) ):
            (match, time_diff, tec1, tec2) = match_up_TEC_obs(aux_tecs1, aux_tecs2[i2])
            tec2_line = aux_tecs2[i2][3].rstrip()
            if ( match and abs(time_diff) < max_time_delta ):
                tec_diff = tec2.val - tec1.val
#                lt = local_time(tec2.time, tec2.lon)
                out_line = join(tec2_line, time_diff, tec1.sat, tec1.az, tec1.el, \
                      tec1.lat, tec1.lon, tec1.val, tec1.err, tec_diff)
                print out_line
                if stats_file != "":
                    gstats.addData(out_line)
            else:
                print '#', tec2_line, time_diff, tec1.sat, tec1.site, "**No_match**"

        tecs2 = []
        tecs2.append(line)
        cur_site = fields[1]
    line = in2.readline()
    while (line and line[0] == '#'): line = in2.readline()

stats = open(stats_file, "w")
for line in gstats.out_lines:
    stats.write(line + '\n')

